/******************************************************************************
Project:		LCMeter
File:			key.c
Author:			Philipp Burch
Date:			01.03.2008
Description:	This code handles the events generated by attached buttons.
				Debouncing is done with Peter Danneggers code, thanks a lot
				for this!
Menu:			The menu used to control the tool is quite simple. In the
				"home screen", components can be measured by selecting the
				desired "target" using the [MODE] button and hitting [START].
				With [CAL], the device measures the built-in capacitor and
				inductor. [+] and [-] have no function.
				With the [M+] and [M-] button, one can navigate through the
				menu. After "home", there follows a display of the built-in
				capacitor and inductor values and finally of the Cref
				capacitor. By hitting [M+] once more, the value of Cref can
				be adjusted with the [+] and [-] buttons. [MODE], [CAL] and
				[START] have are not needed in these menues.
				Possible [MODE]s are:
				 - Inductance (Lx=)
				 - Capacitance (Cx=)
				 - Resonance frequency of an LC-Circuit (LCf=)
				 - Inductance of an LC-Circuit (LCl=)
				 - Capacitance of an LC-Circuit (LCc=)
******************************************************************************/

#include <key.h>
#include <stdint.h>
#include <main.h>
#include <lc.h>
#include <lcd.h>
#include <avr/io.h>
#include <avr/pgmspace.h>

menu_t Current_Menu = MENU_HOME;

//Must be called from the ISR (interrupts deactivated!)
uint8_t key_update() {
	static uint8_t key_state = 0;
	static uint8_t ct0, ct1, rpt;
	uint8_t i;

	i = key_state ^ ~RPI_KEYS;
	ct0 = ~(ct0 & i);
	ct1 = ct0 ^ (ct1 & i);
	i &= ct0 & ct1;
	key_state ^= i;
	i &= key_state;

	if (!(key_state & KEY_RPT)) {
		rpt = KEY_RPT_ST;
	}
	if (!--rpt) {
		rpt = KEY_RPT_NX;
		i |= key_state & KEY_RPT;
	}

	return i;
}

void key_process(uint8_t key_press) {
	if (Current_Menu == MENU_HOME) {
		//Allow [MODE], [CAL] and [START]
		if (key_press & _BV(PIN_KEY_MODE)) {
			lc_switch_mode();
		} else if (key_press & _BV(PIN_KEY_CAL)) {
			lc_calibrate(CS_FIRSTCALL);
		} else if (key_press & _BV(PIN_KEY_START)) {
			lc_measure(CS_FIRSTCALL);
		}
	} else if (Current_Menu >= MENU_EDIT_CREF1) {
		//Allow [+] and [-]
		if (key_press & _BV(PIN_KEY_PLUS)) {
			lc_cref_edit(Current_Menu - MENU_EDIT_CREF1 + 1);
		} else if (key_press & _BV(PIN_KEY_MINUS)) {
			lc_cref_edit(-(Current_Menu - MENU_EDIT_CREF1 + 1));
		}
	}
	//[M+] and [M-] are allowed everywhere
	if (key_press & _BV(PIN_KEY_MPLUS)) {
		if (Current_Menu < MENU_EDIT_CREF6) {
			++Current_Menu;
		} else {
			Current_Menu = MENU_HOME;	//Wrap-around
		}
		key_update_menu();
	} else if (key_press & _BV(PIN_KEY_MMINUS)) {
		if (Current_Menu > MENU_HOME) {
			--Current_Menu;
		} else {
			Current_Menu = MENU_SHOW_CREF;	//Wrap-around to the DISPLAY of Cref
		}
		key_update_menu();
	}
}

void key_update_menu() {
	if (Current_Menu < MENU_EDIT_CREF1) lcd_setcursor(16);

	switch (Current_Menu) {
		case MENU_HOME:
			lc_show_home();
			break;
		case MENU_SHOW_L:
			lc_show_l();
			break;
		case MENU_SHOW_C:
			lc_show_c();
			break;
		case MENU_SHOW_CREF:
			lc_show_cref(false);
			break;
		default:
			//Edit cref
			lc_show_cref(true);
			lcd_setcursor(5 + 5 - (Current_Menu - MENU_EDIT_CREF1));
	}
}
