/******************************************************************************
Project:		LCMeter
File:			lcd.c
Author:			Philipp Burch
Date:			29.02.2008
Description:	This file contains the code to take the actual measurements of
				inductance and capacitance.
******************************************************************************/


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <stdint.h>
#include <string.h>
#include <lc.h>
#include <main.h>
#include <lcd.h>


#define CTRL_MASK		(_BV(PIN_CTRL_CREF) | \
						 _BV(PIN_CTRL_C) | \
						 _BV(PIN_CTRL_LP) | \
						 _BV(PIN_CTRL_LS) | \
						 _BV(PIN_CTRL_EN))

#define P_CREF			PIN_CTRL_CREF
#define P_C				PIN_CTRL_C
#define P_LP			PIN_CTRL_LP
#define P_LS			PIN_CTRL_LS
#define P_EN			PIN_CTRL_EN

#define LCC_OPEN		(0)
#define LCC_CREF_L		(_BV(P_EN) | _BV(P_CREF) | _BV(P_LP))
#define LCC_C_L			(_BV(P_EN) | _BV(P_C) | _BV(P_LP))
#define LCC_CREF_C_L	(_BV(P_EN) | _BV(P_CREF) | _BV(P_C) | _BV(P_LP))
#define LCC_LX_CREF		(_BV(P_EN) | _BV(P_CREF))
#define LCC_LX_CREF_C	(_BV(P_EN) | _BV(P_CREF) | _BV(P_C))
#define LCC_LXP_CREF	(_BV(P_EN) | _BV(P_LP) | _BV(P_CREF))
#define LCC_LXS_CREF	(_BV(P_LS) | _BV(P_CREF))
#define LCC_CX_L		(_BV(P_EN) | _BV(P_LP) | _BV(PC6))
#define LCC_CX_CREF_L	(_BV(P_EN) | _BV(P_CREF) | _BV(P_LP) | _BV(PC6))
#define LCC_CX_CREF_C_L	(_BV(P_EN) | _BV(P_CREF) | _BV(P_C) | _BV(P_LP))
#define LCC_LC			(_BV(P_EN))
#define LCC_LC_CREF		(_BV(P_EN) | _BV(P_CREF) | _BV(PC6))

#define T_SETTLE		500
#define F_MAX			1000000.0


uint32_t Cref_val_eep __attribute__((section(".eeprom"))) = 1000UL;	//In pF
uint32_t Cref_val_eep_l;
char Cref_val_str[7];
float Cref_val;
float L_val;
float C_val;
float f_one = 1.0;
volatile uint8_t timer_b3;
float frequency;
mode_t Current_Mode = MODE_L;
uint8_t control = LCC_OPEN;
uint32_t digit_value[] PROGMEM = {100000UL, 10000UL, 1000UL, 100UL, 10UL, 1UL};

prog_char sLxIs[] PROGMEM = "Lx=";
prog_char sCxIs[] PROGMEM = "Cx=";
prog_char sLCfIs[] PROGMEM = "LCf=";
prog_char sLClIs[] PROGMEM = "LClx=";
prog_char sLCcIs[] PROGMEM = "LCc=";
prog_char sLIs[] PROGMEM = "L=";
prog_char sCIs[] PROGMEM = "C=";
prog_char sCrefIs[] PROGMEM = "Cref=";
prog_char sH[] PROGMEM = "H";
prog_char sF[] PROGMEM = "F";
prog_char sHz[] PROGMEM = "Hz";
prog_char *mode_prefix[] = {sLxIs, sCxIs, sLCfIs, sLClIs, sLCcIs};
prog_char *mode_suffix[] = {sH, sF, sHz, sH, sF};

void lc_control(uint8_t ctrl) {
	control = ctrl;
	RPO_CONTROL |= CTRL_MASK;
	RPO_CONTROL &= ~ctrl;
}

void eep_write_cref() {
	eeprom_write_block(&Cref_val_eep_l, &Cref_val_eep, 4);
}

void eep_read_cref() {
	eeprom_read_block(&Cref_val_eep_l, &Cref_val_eep, 4);
	Cref_val = 1.0E-12 * Cref_val_eep_l;
}

void lc_init() {
	RPO_CONTROL &= ~CTRL_MASK;
	wait_ms(200);
	lc_control(LCC_OPEN);

	TIMSK |= _BV(TOIE1);

	eep_read_cref();
	uint32_t temp = Cref_val_eep_l;
	uint32_t v;
	uint8_t num;
	for (uint8_t i = 0; i < 5; ++i) {
		num = 0;
		v = pgm_read_dword(&digit_value[i]);
		while (temp >= v) {
			temp -= v;
			++num;
		}
		Cref_val_str[i] = '0' + num;
	}
	Cref_val_str[5] = '0' + (uint8_t)temp;
	Cref_val_str[6] = 0;
}

void lc_switch_mode() {
	if (Current_Mode < MODE_LC_C) {
		++Current_Mode;
	} else {
		Current_Mode = MODE_L;
	}

	lc_show_home();

	switch (Current_Mode) {
		case MODE_L:
		case MODE_LC_L:
			DDR_LX_CX |= _BV(PIN_LX_CX);
			RPO_LX_CX &= ~_BV(PIN_LX_CX);			
			break;
		case MODE_LC_F:
			DDR_LX_CX &= ~_BV(PIN_LX_CX);
			RPO_LX_CX &= ~_BV(PIN_LX_CX);
			break;
		default:
			DDR_LX_CX |= _BV(PIN_LX_CX);
			RPO_LX_CX |= _BV(PIN_LX_CX);

	}
}

float lc_calc(float *divisor) {
	if (frequency == 0.0 || *divisor == 0.0) return 0.0;
	return SQUARE(1 / (PI2 * frequency)) / *divisor;
}

void lc_calibrate(call_src_t source) {
	/*  Calibrating sequence:
		1. Enable Cref and L and calculate L
		2. Enable C and L and calculate C
		3. Enable Cref, C and L and check the result
	*/

	static uint8_t stage = 0;

	if (source == CS_FIRSTCALL) {
		stage = 0;
		lcd_settext_P(PSTR("Calibrating..."));
		GlobalFlags |= _BV(GF_BUSY);
		lc_control(LCC_CREF_L);
		set_event(T_SETTLE, &lc_calibrate);
	} else {
		switch (stage) {
			case 0:
			case 2:
			case 4:
				lc_get_frequency(CS_FIRSTCALL);
				set_event(150, &lc_calibrate);
				break;
			case 1:
				L_val = lc_calc(&Cref_val);
				lc_control(LCC_C_L);
				set_event(T_SETTLE, &lc_calibrate);
				break;
			//case 2: see above
			case 3:
				C_val = lc_calc(&L_val);
				lc_control(LCC_CREF_C_L);
				set_event(T_SETTLE, &lc_calibrate);
				break;
			//case 4: see above
			case 5:
				lc_control(LCC_OPEN);
				//Calculate the deviation from Cref
				float d;
				d = lc_calc(&L_val) - C_val - Cref_val;
				if (d < 0.0) d = -d;
				lcd_printval(PSTR("dCref="), d, PSTR("F"), 4);
				GlobalFlags &= ~_BV(GF_BUSY);
				break;
		}
		++stage;
	}
}

void lc_measure(call_src_t source) {
	/*	Measuring sequence:
		--- Lx ---
		1. Enable Cref and Lp and calculate Lx
		2. If f >= F_MAX, enable Cref and calculate Lx
		4. If f >= F_MAX, enable Cref, C and Ls and calculate Lx
		--- Cx ---
		1. Enable Lp and calculate Cx
		2. If f >= F_MAX, enable Cref and Lp and calculate Cx
		3. If f >= F_MAX, enable Cref, Lp and C and calculate Cx
		--- LC (frequency) ---
		1. Just measure the frequency
		--- LCl (inductive component of an LC-circuit) ---
		1. Measure the circuit's resonance frequency
		2. Enable Cref and calculate the capacitive component
		3. Calculate the inductive component
		--- LCc (capacitive component of an LC-circuit) ---
		Same as for LCl.
	*/

	static uint8_t stage = 0;
	static float x, y;
	static uint8_t mask[5][4] PROGMEM = {
		{LCC_LXP_CREF, LCC_LX_CREF, LCC_LXS_CREF},
		{LCC_CX_L, LCC_CX_CREF_L, LCC_CX_CREF_C_L},
		{LCC_LC, 0, 0},
		{LCC_LC, LCC_LC_CREF, 0},
		{LCC_LC, LCC_LC_CREF, 0}
	};

	if (source == CS_FIRSTCALL) {
		x = 0.0;
		stage = 0;
		GlobalFlags = _BV(GF_BUSY) | (GlobalFlags & ~_BV(GF_NOFREQ));
	}

	if ((stage & 1) == 0) {
		//Even values -> Calculate and switch relays
		if (!(GlobalFlags & _BV(GF_NOFREQ))) {
			if (stage) {
				switch (Current_Mode) {
					case MODE_L:
						if (stage == 2) {
							if (frequency < F_MAX) {
								x = 1.0 / (1.0 / lc_calc(&Cref_val) - 1.0 / L_val);
								stage = 0xff;
							}
						} else if (stage == 4) {
							if (frequency < F_MAX) {
								x = lc_calc(&Cref_val);
								stage = 0xff;
							}
						} else {
							x = lc_calc(&Cref_val) - L_val;
							stage = 0xff;
						}
						break;

					case MODE_C:
						if (stage == 2) {
							if (frequency < F_MAX) {
								x = lc_calc(&L_val);
								stage = 0xff;
							}
						} else if (stage == 4) {
							if (frequency < F_MAX) {
								x = lc_calc(&L_val) - Cref_val;
								stage = 0xff;
							}
						} else {
							x = lc_calc(&L_val) - Cref_val - C_val;
							stage = 0xff;
						}
						break;

					case MODE_LC_F:
						x = frequency;
						stage = 0xff;
						break;

					case MODE_LC_L:
					case MODE_LC_C:
						if (stage == 2) {
							y = lc_calc(&f_one);
						} else {
							stage = 0xff;
							x = Cref_val * y / (lc_calc(&f_one) - y);
							if (Current_Mode == MODE_LC_L) {
								y = x + Cref_val;
								x = lc_calc(&y);
							}
						}
				}
			}
		} else {
			stage = 0xff;
		}

		if (stage == 0xff) {
			lc_control(LCC_OPEN);
			lcd_printval(mode_prefix[Current_Mode], x, mode_suffix[Current_Mode], 4);
			GlobalFlags &= ~_BV(GF_BUSY);
			return;
		}
		lc_control(pgm_read_byte(&mask[Current_Mode][stage / 2]));
		set_event(T_SETTLE, &lc_measure);
	} else {
		//Odd values -> get frequency
		lc_get_frequency(CS_FIRSTCALL);
		set_event(150, &lc_measure);
	}
	++stage;
}

void lc_cref_edit(int8_t inc) {
	uint8_t dig;
	if (inc < 0) {
		dig = -inc;
	} else {
		dig = inc;
	}
	--dig;

	if (inc > 0) {
		if (Cref_val_str[5 - dig] < '9') {
			inc = 1;
		} else {
			inc = -9;
		}
	} else {
		if (Cref_val_str[5 - dig] > '0') {
			inc = -1;
		} else {
			inc = 9;
		}
	}
	Cref_val_str[5 - dig] += inc;
	Cref_val_eep_l += inc * pgm_read_dword(&digit_value[5 - dig]);

	eep_write_cref();
	eep_read_cref();	//Just to get Cref_val right
	lc_show_cref(true);
}

void lc_show_home() {
	lcd_settext_P(mode_prefix[Current_Mode]);
	lcd_buffer()[strlen_P(mode_prefix[Current_Mode])] = '?';
}

void lc_show_l() {
	lcd_printval(sLIs, L_val, sH, 4);
}

void lc_show_c() {
	lcd_printval(sCIs, C_val, sF, 4);
}

void lc_show_cref(bool edit) {
	if (!edit) {
		lcd_printval(sCrefIs, Cref_val, sF, 4);
	} else {
		char *s = lcd_buffer();
		strcpy_P(s, sCrefIs);
		strcpy(s + 5, Cref_val_str);
		s[11] = 'p';
		s[12] = 'F';
		for (uint8_t i = 13; i < 16; ++i) s[i] = ' ';
	}
}

void lc_get_frequency(call_src_t source) {
	if (source == CS_FIRSTCALL) {
		timer_b3 = 0;
		frequency = 0.0;
		GlobalFlags |= _BV(GF_GETFREQ);
	} else {
		if (timer_b3 || TCNT1 > 1) {
			frequency = ((float)(((int32_t)timer_b3 << 16) | TCNT1)) * 5.0;
			GlobalFlags &= ~_BV(GF_NOFREQ);
		} else {
			GlobalFlags |= _BV(GF_NOFREQ);
		}
	}
}


ISR(TIMER1_OVF_vect) {
	++timer_b3;
}
